/*
**	Programma:		SampleServer
**	Versione:		1.0.0
**	Autore:		Dario de Judicibus
**	Indirizzo:		dejudicibus@geocities.com
**	Creato il:		January 3rd, 1998
**	Modificato il:	January 4th, 1998
**	Linguaggio:		Java (jdk 1.1)
**	-------------------------------------------------------------
**	(c) 1998 - All rights reserved 
**	-------------------------------------------------------------
**	I make no representations or warranties about the suitability of
**	the software, either express or implied, including but not limited
**	to the implied warranties of merchantability, fitness for a
**	particular purpose, or non-infringement. I shall not be liable for
**	any damages suffered by licensee as a result of using, modifying or
**	distributing this software or its derivatives.
*/

/**
 *	@author Dario de Judicibus
 *	@version 1.0.0
 *	@see jdk 1.1
 */
import java.io.* ;
import java.net.* ;
import java.util.Date ;
import java.text.SimpleDateFormat ;

class SampleServer
{
	// CLASS VARIABLES
	private static final int cvPORT = 2000 ;
	private static final String 	cvDATEFORMAT = "EEE dd MMM yyyy HH:mm:ss z" ;

	// INSTANCE VARIABLES
	private String 			ivServerName ;
	private ServerSocket		ivServerSocket ;
	private Socket			ivClientSocket ;
	private BufferedReader		ivFromClientStream ;
	private PrintWriter		ivToClientStream ;
	private ServerFSM			ivServerFSM ;
	private InetAddress		ivClientAddress ;
	private SimpleDateFormat	ivFormatter ;

	public static void main( String[] args )
	{
		SampleServer sampleServer = new SampleServer( ) ;
		sampleServer.run( ) ;
	}

	private SampleServer( )
	{
		ivFormatter = new SimpleDateFormat( cvDATEFORMAT ) ;

		System.out.println( "[" + ivFormatter.format( new Date( ) ) + "] " + "Setting up server." ) ; 
	}

	private void run( )
	{
		// Try to create a server socket first.
		// If fails, just abend the server.
		try
		{
      		createServerSocket( ) ;
		}
		catch ( IOException exc )
		{
			System.out.println( "[ERROR] " + exc.getMessage( ) ) ;
			System.exit( 1 ) ; // Terminate server
		}
	
		// Try to accept a connection from client, now.
		// If fails, just abend the server.
        	try 
		{
        		acceptClientConnection( ) ;
        	} 
		catch ( IOException exc ) 
		{
			System.out.println( "[ERROR] " + exc.getMessage( ) ) ;
			System.exit( 1 ) ; // Terminate server
        	}

		// Manage interactions with client.
		// If fails, print the stack trace and abend the server.
		try
		{
			// Prepare for data exchange
			prepareForExchange( ) ;

			String inputLine, outputLine ;

			// Server introduces itself
			outputLine = ivServerFSM.processInput( null ) ;
			ivToClientStream.println( outputLine ) ;

			long skipped ;

			// Process input from server until FSM returns "null"
			while ( ( inputLine = ivFromClientStream.readLine( ) ) != null )
			{
				// Skip termination character(s)
				// Say '2' since they can be either of
				// <NL>, <CR>, <CRLF>, <0x00>
				skipped = ivFromClientStream.skip( 2 ) ;

				System.out.println( "[" + ivFormatter.format( new Date( ) ) + "] " 
					+ "C: {" + inputLine + "}" ) ;  

				outputLine = ivServerFSM.processInput( inputLine ) ;
				if ( outputLine == null )
                    		break ;

				ivToClientStream.println( outputLine ) ;
				System.out.println( "[" + ivFormatter.format( new Date( ) ) + "] " 
					+ "S: {" + outputLine + "}" ) ;  
			}

			// Terminate session and close connection
			terminateConnection( ) ;

		} 
		catch ( IOException exc )
		{
			exc.printStackTrace( ) ;
		}
	}

	private void createServerSocket( ) throws IOException
	{
		ivServerSocket = new ServerSocket( cvPORT ) ;

		ivServerName = InetAddress.getByName( null ).getHostName( ) ;

		System.out.println( "[" + ivFormatter.format( new Date( ) ) + "] " 
			+ "Listening on port " + cvPORT ) ;      

		return ;
	}

	private void acceptClientConnection( ) throws IOException
	{
           	ivClientSocket = ivServerSocket.accept( ) ;
		ivClientAddress = ivClientSocket.getInetAddress( ) ;

		System.out.println( "[" + ivFormatter.format( new Date( ) ) + "] "
			+ "Accepted connection from " + ivClientAddress ) ;       

		return ;
	}

	private void prepareForExchange( ) throws IOException
	{
		// Create the input and output streams
		// Both streams are buffered
           	ivFromClientStream = new BufferedReader( 
			new InputStreamReader( ivClientSocket.getInputStream( ) ) ) ;

           	ivToClientStream = new PrintWriter(	ivClientSocket.getOutputStream( ), true ) ;

		// Prepare the Finite State Machine for protocol handling
		ivServerFSM = new ServerFSM ( ivServerName ) ;     

		System.out.println( "[" + ivFormatter.format( new Date( ) ) + "] " + "Server ready" ) ; 

		return ;
	}

	private void terminateConnection( ) throws IOException
	{
            ivToClientStream.close( ) ;
            ivFromClientStream.close( ) ;
            ivClientSocket.close( ) ;
            ivServerSocket.close( ) ;

		System.out.println( "[" + ivFormatter.format( new Date( ) ) + "] "
			+ "Closing connection from " + ivClientAddress ) ;

		// Reset all instance variables
		ivClientAddress = null ;       
            ivToClientStream = null ;       
            ivFromClientStream = null ;       
            ivClientSocket = null ;       
            ivServerSocket = null ;       

		return ;
	}
}

class ServerFSM 
{
	// CLASS VARIABLES
	private static final int cvIDLE 		= 0 ;
	private static final int cvCONNECTED 	= 1 ;
	private static final int cvLOGGED 		= 2 ;
	private static final int cvCLOSING 		= 9 ;

	// INSTANCE VARIABLES
	private int ivState = cvIDLE ;
	private String ivServer ;

	ServerFSM( String server )
	{
		ivServer = server ;
	}

	String processInput( String inputString ) 
	{
      	String outputString = null ;

		if ( ivState == cvIDLE ) 
		{
			if ( inputString == null )
			{
            		outputString = "OK Server " + ivServer ;
				ivState = cvCONNECTED ;
			}
			else if ( inputString.equals( "LOGOFF" ) )
			{
            		outputString = "OK Closing server." ;
				ivState = cvCLOSING ;
			}
			else
			{
            		outputString = "BAD Wait for greetings." ;
			}
		}
		else if ( ivState == cvCONNECTED )
		{
			if ( inputString.equals( "LOGIN" ) )
			{
            		outputString = "OK You are logged in." ;
				ivState = cvLOGGED ;
			}
			else if ( inputString.equals( "LOGOFF" ) )
			{
            		outputString = "OK Closing server." ;
				ivState = cvCLOSING ;
			}
			else
			{
            		outputString = "BAD You must login, first." ;
			}
		}
		else if ( ivState == cvLOGGED )
		{
			if ( inputString.equals( "LOGOFF" ) )
			{
            		outputString = "OK Closing server." ;
				ivState = cvCLOSING ;
			}
			else
			{
            		outputString = "BAD You can just logoff, now." ;
			}
		}
		else if ( ivState == cvCLOSING )
		{
           		outputString = null ;
			ivState = cvIDLE ;
		}

		return outputString ;
	}
}
